@@ -115,6 +115,7 @@ class BertTagger:
     def set_models_dir(self, model_dir: str):
         self.models_dir = model_dir
 
+
     def _to_megabytes(self, _bytes: int) -> int:
         """ Convert bytes to megabytes.
 
@@ -167,40 +168,46 @@ class BertTagger:
         """ Get device to run the process on (GPU, if available, else CPU).
         """
         device = None
+        use_cpu = True
         cpu_free_memory = self._get_free_memory_cpu()
         # If there's a GPU available...
         if self.use_gpu and torch.cuda.is_available():
 
             # Check if available GPU(s) have required amount of available memory.
             device_ids = GPUtil.getAvailable(order ="memory", maxMemory = self.gpu_max_memory)
-            device_id, device_memory_info = self._sort_by_memory(device_ids, "free")[0]
+            try:
+                device_id, device_memory_info = self._sort_by_memory(device_ids, "free")[0]
+                use_cpu = False
 
-            if device_memory_info["free"] >= self.min_available_memory or cpu_free_memory < self.min_available_memory:
-                torch.cuda.empty_cache()
-                device = torch.device("cuda")
-                torch.cuda.set_device(device_id)
-                self._print(f"We will use the GPU: {torch.cuda.get_device_name(device_id)}")
-                self._print(f"Device memory: [Free: {device_memory_info['free']} MB; Used: {device_memory_info['used']} MB; Total: {device_memory_info['total']} MB]")
-                self._print(f"Device index: {torch.cuda.current_device()}")
-                if self.logger:
-                    self.logger.info(f"We will use the GPU: {torch.cuda.get_device_name(device_id)}")
-                    self.logger.info(f"Device memory: [Free: {device_memory_info['free']} MB; Used: {device_memory_info['used']} MB; Total: {device_memory_info['total']} MB]")
-                    self.logger.info(f"Device index: {torch.cuda.current_device()}")
-                self.device_available_memory = device_memory_info["free"]
-                self.device_type = "gpu"
-            else:
-                self._print(f"No GPU available with free memory >= {self.min_available_memory}. Using CPU.")
-                if self.logger:
-                    self.logger.info(f"No GPU available with free memory >= {self.min_available_memory}. Using CPU.")
-                device = torch.device("cpu")
-                self.device_type = "cpu"
-                self.device_available_memory = cpu_free_memory
+                if device_memory_info["free"] >= self.min_available_memory or cpu_free_memory < self.min_available_memory:
+                    torch.cuda.empty_cache()
+                    device = torch.device("cuda")
+                    torch.cuda.set_device(device_id)
+                    self._print(f"We will use the GPU: {torch.cuda.get_device_name(device_id)}")
+                    self._print(f"Device memory: [Free: {device_memory_info['free']} MB; Used: {device_memory_info['used']} MB; Total: {device_memory_info['total']} MB]")
+                    self._print(f"Device index: {torch.cuda.current_device()}")
+                    if self.logger:
+                        self.logger.info(f"We will use the GPU: {torch.cuda.get_device_name(device_id)}")
+                        self.logger.info(f"Device memory: [Free: {device_memory_info['free']} MB; Used: {device_memory_info['used']} MB; Total: {device_memory_info['total']} MB]")
+                        self.logger.info(f"Device index: {torch.cuda.current_device()}")
+                    self.device_available_memory = device_memory_info["free"]
+                    self.device_type = "gpu"
+                else:
+                    self._print(f"No GPU available with free memory >= {self.min_available_memory}. Using CPU.")
+                    if self.logger:
+                        self.logger.info(f"No GPU available with free memory >= {self.min_available_memory}. Using CPU.")
+                    device = torch.device("cpu")
+                    self.device_type = "cpu"
+                    self.device_available_memory = cpu_free_memory
+            except:
+                use_cpu = True
 
-        else:
-            self._print("No GPU available or GPU usage turned off, using the CPU instead.")
+
+        if use_cpu:
+            self._print("No GPU available, not enough GPU memory, or GPU usage turned off, using the CPU instead.")
             self._print(f"Available memory: {cpu_free_memory} MB")
             if self.logger:
-                self.logger.info("No GPU available or GPU usage turned off, using the CPU instead.")
+                self.logger.info("No GPU available, not enough GPU memory, or GPU usage turned off, using the CPU instead.")
                 self.logger.info(f"Available memory: {cpu_free_memory} MB")
 
             device = torch.device("cpu")
@@ -623,6 +630,20 @@ class BertTagger:
         return (input_ids, attention_masks, labels)
 
 
+    def _force_str(self, data_sample: Dict[str, list]) -> Dict[str, List[str]]:
+        """ Force string on each value.
+
+        Parameters:
+            data_sample: Training (and valiation) data as dict where keys = labels, values = example texts.
+
+        """
+        self._print("Forcing all values to string...")
+        if self.logger: self.logger.info("Forcing all values to string...")
+        for key, values in list(data_sample.items()):
+            data_sample[key] = [str(v) for v in values]
+        return data_sample
+
+
     def _get_dataset(self, input_ids, attention_masks, labels=[]) -> TensorDataset:
         """ Combine the training inputs into a TensorDataset.
 
@@ -721,7 +742,7 @@ class BertTagger:
         """ Extract examples and labels from data_samples.
 
         Parameters:
-            data_sample - TODO
+            data_sample - Training (and valiation) data as dict where keys = labels, values = example texts.
 
         Returns:
             examples - Training (and validation) examples.
@@ -753,6 +774,9 @@ class BertTagger:
         if self.logger: self.logger.info("Preparing_data...")
         self.config.n_classes = len(data_sample)
 
+
+        data_sample = self._force_str(data_sample)
+
         self._create_label_indices(data_sample)
         self.config.classes = list(self.config.label_index.keys())
         self.config.class_indices = list(self.config.label_index.values())
@@ -810,14 +834,16 @@ class BertTagger:
         torch.cuda.manual_seed_all(self.config.seed_val)
 
 
-    def train(self, data_sample: Dict[str, List[str]], **kwargs) -> TaggingReport:
+    def train(self, data_sample: Dict[str, List[str]], from_checkpoint: bool = False, **kwargs) -> TaggingReport:
         """ Fine-tune BERT model.
 
         Parameters:
             data_sample - Training (and valiation) data as dict where
                           keys = labels, values = example texts.
+            from_checkpoint - If enabled, use a loaded fine-tuned model.
 
         """
+        
         InputValidator().validate(data_sample, **kwargs)
 
         self.update_config(**kwargs)
@@ -833,7 +859,18 @@ class BertTagger:
         total_t0 = time.time()
         self.epoch_reports = []
 
-        self.model = self.load_pretrained()
+
+        # If from_checkpoint is enabled, check if the model to fine-tune actually exists
+        if from_checkpoint:
+            if self.model:
+                self._print("Continuing training from a checkpoint...")
+                if self.logger: self.logger.info(f"Continuing training from a checkpoint...")
+            else:
+                raise exceptions.ModelNotLoadedError(f"No model found. You should load a fine-tuned model or train a model with 'from_checkpoint' set to False.")
+
+        # If from checkpoint is disabled, load a fresh state of a pre-trained model.
+        else:
+            self.model = self.load_pretrained()
 
         optimizer = self.get_optimizer()
         scheduler = self.get_scheduler(n_training_batches, optimizer)
@@ -938,7 +975,7 @@ class BertTagger:
         """
         combined_text = []
         for field_content in list(doc.values()):
-            combined_text.append(field_content)
+            combined_text.append(str(field_content))
         combined_text = " ".join(combined_text)
         predicted_item = self.tag_text(combined_text)
         return predicted_item
